#!/usr/local/bin/perl -w
#############################################################################
#
#  Program:   local_normal.pl
#  Function:  To normalise micro-array data using the method of
#             R. Lyne et al. (2003), BMC Genomics 2003, 4:27.
#             *Heavily* borrows code from the normalisation scripts of
#             Juan Mata and Rachel Lyne.
#
#  Author:    Chris Penkett
#  Date:      6 May 2003
#  Address:   The Wellcome Trust Sanger Institute
#             Wellcome Trust Genome Campus
#             Hinxton
#             Cambridge
#             CB10 1SA
#  Telephone: +44 (0)1223 494862
#  Email:     cjp@sanger.ac.uk
#
#############################################################################
#
#  Description:
#
#:  This script normalises micro-array data using the method described in
#:  the following paper:
#:    Rachel Lyne, Gavin Burns, Juan Mata, Chris J Penkett, Gabriella
#:    Rustici, Dongrong Chen, Cordelia Langford, David Vetrie, Jurg Bahler;
#:    BMC Genomics 2003, 4:27 (10 July 2003).
#:    Whole-genome microarrays of fission yeast: characteristics, accuracy,
#:    reproducibility, and processing of array data.
#:
#:  If you use this script, please cite the above reference.
#:
#:  *Heavily* borrows code from the normalisation scripts of
#:  Juan Mata and Rachel Lyne.
#:
#:  This perl script uses GenePix results files as input and discards data
#:  from spots with failed or faulty PCR products by masking them absent.
#:  Data from spots with low array element concentration (as judged by PCR
#:  product staining on gel) or PCR products where the reverse primer is
#:  located 2500-3500 bp from the gene end are flagged marginal.
#:  In the GenePix results file, the 'ID' column is used for the gene
#:  ID and the 'Name' column is used for information about the type or
#:  quality of the PCR spot on the micro-array slide.  In the 'Name' column,
#:  failed or faulty spots are set to 'A' (absent), and spots with low probe
#:  concentration or where the reverse primer is located 2500-3000 bp from
#:  the 3' end of the gene are set to 'M' (marginal).  Also, in the 'Name'
#:  column, 'Q' indicates bacterial controls, 'T' indicates tags or markers,
#:  and 'Y' indicates genes that are controls for cross-hybrididation
#:  (budding yeast genes of different levels of homology in our case).
#:
#:  This information is then used together with the 'Flags' column in the
#:  results file regarding the spot finding algorithm.  For example, spots
#:  not found are given a flag of '-50', and spots masked as bad by the user
#:  are given a flag of '-100' by GenePix.  These flags are masked as being
#:  absent ('A').  For each spot, a combined flag is produced such that any
#:  spot that is absent in either the 'Name' column or the 'Flags' column is
#:  given an overall flag of absent ('A').  The combined flag is set to
#:  marginal ('M') for any spot that is marginal in the 'Name' column.  This
#:  option is performed automatically on all data.
#:
#:  The script applies cut-off criteria to discard data from weak signals:
#:  spots with < 50% of pixels > 2 SD above median local background signal in
#:  one or both channels are flagged absent, unless one channel shows > 95%
#:  of the pixels > 2 SD above local background.  The SD was calculated using
#:  only the lower 55% of the pixel intensities (called SD2 in GenePix Pro),
#:  as this measure is less susceptible to being skewed by bright pixels.
#:  The default cut-offs of 50% and 95% can be changed in the GUI, and this
#:  filtering option can also be switched off.  The script alsoe provides a
#:  quality control report showing the numbers of spots discarded during the
#:  various steps of the data analysis pipeline as well as information about
#:  genes with % CVs of signal ratios > 30% between replicate spots.
#:
#:  The script also performs a local normalisation using a sliding square
#:  window of spots surrounding each spot.  A user-defined minimum number of
#:  spots is chosen to be used with which to normalise over (default is 400).
#:  The window size default is 16 spots.  This means the square contains
#:  33 x 33 spots (1089) surrounding central spots,
#:  33 x 17 spots (561) surrounding spots at the edge of the array, and
#:  17 x 17 spots (289) surrounding spots in a corner of the array.  Only
#:  spots that are flagged present are used for the normalisation.  Hence,
#:  using a window of 16 means that sometimes, especially for spots close
#:  to the corners of the array, less than 400 spots may actually be used for
#:  the normalisation.  In cases where the block size chosen is small, the
#:  window size is increased up to a user-defined maximum window size
#:  (default is 24) so that at least 600 total spots are in the square.  This
#:  means the block size used with this window change is larger than may be
#:  necessary to optimise the chances of having 400 present spots to use for
#:  normalisation.  This is a heuristic to make the algorithm faster for the
#:  majority of spots, since counting the number of present spots in the
#:  initial square uses a relatively large amount of computational time.  If,
#:  during normalisation, the number of spots is still found to be less than
#:  400, the window is increased further until the maximum window size is
#:  reached.  In these cases, the spots that do use less than 400 spots for
#:  normalisation are reported in the output log file.  The script then
#:  calculates a normalisation factor such that the median signal ratio of
#:  all measurable spots within the square equals 1, and this factor is then
#:  used to scale the signal ratio for the central spot.
#:
#:  The signal ratios used for normalisation correspond to the median of all
#:  pixel-by-pixel ratios of signals minus median local background for each
#:  pixel of a given spot (called 'median of ratios' in GenePix Pro).  This
#:  measures ratios more reliably and is less affected by unspecific signals
#:  than the 'ratio of medians'.  In the rare cases where the
#:  'median of ratios' is zero, the 'ratio of medians' is used instead for
#:  data evaluation.  Finally, the script averages the normalised data from
#:  all replicate spots that produced measurable signal ratios of the same
#:  genomic element.  These mean normalised ratios can then be used for
#:  downstream data evaluation.
#:
#############################################################################
#
#  Usage:
#
#:  local_normal.pl
#:   [--help                            ]
#:   [--os          <Boolean>           ]
#:   [--file        <File>              ]
#:   [--genefile    <File>              ]
#:   [--extgene     <Boolean>           ]
#:   [--bactcont    <Boolean>           ]
#:   [--window      <Number>            ]
#:   [--maxwin      <Number>            ]
#:   [--blocksize   <Number>            ]
#:   [--blockstart  <Number>            ]
#:   [--bgcut       <Boolean>           ]
#:   [--upperbg     <Number>            ]
#:   [--lowerbg     <Number>            ]
#:   [--scaling     <Number>            ]
#:   [--diamcut     <Number|Boolean>    ]
#:   [--spotnorm    <Boolean>           ]
#:   [--genenorm    <Boolean>           ]
#:   [--genestats   <Boolean>           ]
#:   [--alldata     <Boolean>           ]
#:   [--log         <Boolean>           ]
#:   [--info        <Boolean>           ]
#:
#:    --help        Shows usage.
#:    --os          OS information;
#:                    Windows is true (1), UNIX is false (0; default).
#:    --file        File(s) to be normalised.
#:    --genefile    Gene list file.
#:    --extgene     Use external gene list from separate file;
#:                    (default is gene_list.txt in same directory as script).
#:                    If false, use gene list made from input data file.
#:    --bactcont    Calculate mean ratio of bacterial controls
#:                    (default is false); and use for normalisation.
#:    --window      Size of normalisation window (default is 16).
#:    --maxwin      Maximum size of normalisation window (default is 24).
#:    --blocksize   Size of normalisation block (default is 400).
#:    --blockstart  Initial size of normalisation block (default is 600).
#:                    This is used to account for absent spots
#:                    in the initial square used.
#:    --bgcut       Use background SD cut-offs (default is true).
#:    --upperbg     Upper cut-off for background SD (default is 95).
#:    --lowerbg     Lower cut-off for background SD (default is 50).
#:    --scaling     Multiplification factor for use when ratios are not
#:                    expected to be 1.0 (e.g., CGH during cell cycle).
#:    --diamcut     Use/set spot diameter cut-off (default is 100);
#:                    default is not to use this cut-off.
#:    --spotnorm    Calculate normalised data for all good spots.
#:    --genenorm    Calculate normalised data for all good genes.
#:    --genestats   Prints extra information about normalised data
#:                    for all good genes (default is false).
#:    --alldata     Prints all data for all spots plus normalised data.
#:    --log         Print log information to a file (default is true);
#:                    otherwise print it to stderr.
#:    --info        Print header information to a file.
#:
#:  If no command line parameters are supplied, a graphical user
#:  interface is started instead.
#:
#############################################################################
#
#  Copyright: (C) 2003  Genome Research Limited.
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the:
#    Free Software Foundation, Inc.
#    59 Temple Place - Suite 330
#    Boston, MA 02111-1307, USA.
#
#############################################################################
#
#  Revision History:
#
#  $Author: cjp $
#  $Date: 2003/08/29 15:16:16 $
#  $Revision: 1.20 $
#  $RCSfile: local_normal.pl,v $
#
#  $Log: local_normal.pl,v $
#% Revision 1.20  2003/08/29  15:16:16  cjp
#% Changed dongle error message.
#%
#% Revision 1.19  2003/08/19  05:41:17  cjp
#% Changes for GenePix 5.
#%
#% Revision 1.18  2003/08/12  17:32:36  cjp
#% Problem with last fix.  Changed back to previous code
#%   and actually fixed original bug.
#%
#% Revision 1.17  2003/08/12  15:27:19  cjp
#% Fixed bacterial control bug for absent spots.
#%
#% Revision 1.16  2003/08/11  15:38:14  cjp
#% Fixed recursion bug.
#%
#% Revision 1.15  2003/08/01  15:14:19  cjp
#% Fixed bugs with flags.
#%
#% Revision 1.14  2003/08/01  10:46:42  cjp
#% Various bug fixes.  Also changed the way controls are handled.
#%
#% Revision 1.13  2003/07/30  15:05:04  cjp
#% Fixed a few bugs.
#%
#% Revision 1.12  2003/07/15  07:30:01  cjp
#% Counting marginal genes correctly now.
#%
#% Revision 1.11  2003/07/14  13:08:57  cjp
#% Slight changes.
#%
#% Revision 1.10  2003/07/11  15:23:58  cjp
#% Bug with --extgene option.
#%
#% Revision 1.9  2003/07/11  15:16:46  cjp
#% Slight changes to description.
#%
#% Revision 1.8  2003/07/11  12:36:51  cjp
#% Added extra description info.
#%
#% Revision 1.7  2003/07/08  16:04:52  cjp
#% Fixed bug when no genes with high CVs between spots.
#%
#% Revision 1.6  2003/07/04  12:49:33  cjp
#% Changes for windows.
#%
#% Revision 1.5  2003/07/04  08:33:55  cjp
#% Output format changes.
#%
#% Revision 1.4  2003/06/10  14:27:28  cjp
#% Debug of GenePix3 code.
#%
#% Revision 1.3  2003/05/27  09:19:22  cjp
#% New algorithm to make sure we normalise over a minimum number of spots.
#%
#% Revision 1.2  2003/05/16  15:57:09  cjp
#% Added Tk GUI frontend.
#%
#% Revision 1.1  2003/05/12  16:29:02  cjp
#% Initial revision.
#%
#% Revision 1.0  2003/05/06  09:00:00  cjp
#% Started writing script.
#
#############################################################################

require 5.004;


#############################################################################
#  Pragmas
#  -------

use strict;


#############################################################################
#  Modules
#  -------

use Cwd;
#use Data::Dumper;
use Getopt::Long;
use Tk;


#############################################################################
#  Constants
#  ---------

# Version information, the revision number changes automatically
#   whenever the script is checked into CVS or RCS.

use constant VERSION => q$Revision: 1.20 $ =~ /[.\d]+/g;

# OS constants.

use constant UNIX    => 0;
use constant WINDOWS => 1;

# Default values.

use constant DEF_CV_CUT => 30;          # Default % CV cut-off for genes
                                        #   with highly variable spots.

use constant DEF_BACT_MIN => 10;        # Default for minimum no. of
                                        #   bacterial controls to use.

use constant DEF_SIGNAL_CUT => 0;       # Default for signal cut-off.

use constant DEF_GENE_LIST_FILE => 'gene_list.txt'; # Default gene list file.

# Default number options.

use constant DEF_WINDOW_SIZE => 16;     # Default for --window option.
use constant DEF_MAXWIN_SIZE => 24;     # Default for --maxwin option.
use constant DEF_BLOCK_SIZE  => 400;    # Default for --blocksize option.
use constant DEF_BLOCKSTART_SIZE => 600;  # Default for --blockstart option.

use constant DEF_UPPER_BG_CUT_SIZE => 95;   # Default for --upperbg option.
use constant DEF_LOWER_BG_CUT_SIZE => 50;   # Default for --lowerbg option.

use constant DEF_SCALE_FACTOR => 1.0;   # Default for --scaling option.

use constant DEF_DIAM_CUT_SIZE => 100;  # Default for diameter cut-off.

# Default boolean options.

use constant DEF_EXT_GENE_LIST => 1;    # Default for --extgene option.
use constant DEF_BACT_CONT_OPT => 0;    # Default for --bactcont option.
use constant DEF_BG_CUT_OPT    => 1;    # Default for --bgcut option.
use constant DEF_DIAM_CUT_OPT  => 0;    # Default for --diamcut option.
use constant DEF_SPOT_NORM_OPT => 1;    # Default for --spotnorm option.
use constant DEF_GENE_NORM_OPT => 1;    # Default for --genenorm option.
use constant DEF_GENE_STAT_OPT => 0;    # Default fro --genestats option.
use constant DEF_ALL_DATA_OPT  => 0;    # Default for --alldata option.
use constant DEF_LOG_FILE_OPT  => 1;    # Default for --log option.
use constant DEF_INFO_FILE_OPT => 0;    # Default for --info option.

# gDataArray indices.

use constant INDEX       => 0;      # Index location on the micro-array.
use constant ARRAY_BLOCK => 1;      # Block in micro-array.
use constant BLOCK_COL   => 2;      # Column in block.
use constant BLOCK_ROW   => 3;      # Row in block.
use constant ARRAY_COL   => 4;      # Column in micro-array.
use constant ARRAY_ROW   => 5;      # Row in micro-array.
use constant RATIO       => 6;      # Ratio from GenePix file.
use constant NORM_FACTOR => 7;      # Normalisation factor.
use constant NORM_RATIO  => 8;      # Normalised ratio.
use constant X_POS       => 9;      # X position on micro-array.
use constant Y_POS       => 10;     # Y position on micro-array.
use constant MASK        => 11;     # Mask (or name) from GenePix file.
use constant ID          => 12;     # ID from GenePix file.
use constant FLAG        => 13;     # Integrated flag from GenePix file
                                    #   and mask information.
use constant GPFLAG      => 14;     # GenePix flag.
use constant GSFLAG      => 15;     # GeneSpring flag.
use constant B635SD2     => 16;     # % of Cy5 signal pixels > BG 635 + 2SD2.
use constant B532SD2     => 17;     # % of Cy3 signal pixels > BG 532 + 2SD2.
use constant F635SAT     => 18;     # % of saturated Cy5 signal pixels.
use constant F532SAT     => 19;     # % of saturated Cy3 signal pixels.
use constant DIAM        => 20;     # Spot diameter.
use constant NO_SPOTS    => 21;     # No. spots to calculate the
                                    #   normalistion factor.
use constant MED_RATIO   => 22;     # Ratio of medians.
use constant RATIO_MED   => 23;     # Median of ratios.
use constant SIG_EXPT    => 24;     # Signal1 (experiment).
use constant SIG_CONT    => 25;     # Signal2 (control).
use constant TRUST       => 26;     # Trust is lowest signal in
                                    #   experiment or control.
use constant AVG_LOG_SIG => 27;     # Average log intensity for both signals.
use constant SIG_RATIO   => 28;     # Ratio of signals.
use constant SIG_NORM_FACTOR => 29; # Normalisation factor of signals.
use constant SIG_NORM_RATIO  => 30; # Normalised ratio of signals.
use constant NORM_SIG_EXPT   => 31; # Normalised signal for experiment.
use constant BACT_NORM_RATIO => 32; # Normalised ratio (of signals)
                                    #   corrected with bact. controls.

# gGeneHash indices.

use constant NO_VAL     => 0;       # No. of values for this gene.
use constant MARG       => 1;       # Flag if this is a marginal gene.
use constant MEAN       => 2;       # Mean value for this gene.
use constant CV         => 3;       # % CV for this gene.
use constant DATA_START => 4;       # Index where normalised ratios begin.


#############################################################################
#  Globals
#  -------

use vars qw(%gOpts $gFileParts
            $gHeader $gDataArray
            @gGeneList $gGeneHash
            $gTk $gDirSep
            $gCyRatio $gGPVers $gBactMean
            $gColsBlock $gRowsBlock $gBlocksArray
            $gBigX $gBigY);

undef %gOpts;       # Hash of options.

$gFileParts = undef;    # Components of file name being processed.
$gHeader    = undef;    # Hash of header information for current file.
$gDataArray = undef;    # Array of data for current file.

undef @gGeneList;   # Array of genes on micro-array;
                    #   either specified in a gene list file
                    #   or in the current file.

$gGeneHash = undef; # Hash of data for genes on micro-array
                    #   in current file.

$gTk = undef;       # Main Tk window.

$gDirSep = '/';     # Directory separator - '/' for UNIX,
                    #   and '\' for Windows.

$gCyRatio = 0;      # Unknown Cy dye ratio.
                    #   1 is 635/532 (Cy5/Cy3); 2 is 532/635 (Cy3/Cy5).

$gGPVers = 3;       # Default GenePix version.

$gBactMean = 0;     # Bacterial controls mean signal ratio.

$gColsBlock   = 0;  # No. of columns in a block on the micro-array.
$gRowsBlock   = 0;  # No. of rows in a block on the micro-array.
$gBlocksArray = 0;  # No. of blocks on the micro-array.

$gBigX = 0;         # Largest possible spot number in X direction.
$gBigY = 0;         # Largest possible spot number in Y direction.


#############################################################################
#  setGlobals
#  ----------

sub setGlobals {

  # These global variables are file-specific;
  #   and should be reset for each new GenePix file.

  $gFileParts = undef;
  $gHeader    = undef;
  $gDataArray = undef;

  $gGeneHash = undef;

  $gCyRatio = 0;

  $gGPVers = 3;

  $gBactMean = 0;

  $gColsBlock   = 0;
  $gRowsBlock   = 0;
  $gBlocksArray = 0;

  $gBigX = 0;
  $gBigY = 0;

  return 1;

} # setGlobals


#############################################################################
#  showUsage
#  ---------

sub showUsage {

  print STDERR
    "\n", (split /$gDirSep/, $0)[-1],
    ' (v ' . VERSION . ")\n\n",
    `perl -ne "s/^#://, print if /^#:/" $0`;

  print STDERR
    "ERROR: ", join(' ', @_), "\n"
      if defined $_[0];

  return 1;

} # showUsage


#############################################################################
#  showExitMessage
#  ---------------

sub showExitMessage {

  my $text = join ' ', @_;

  my $warn = $gTk->Toplevel();

  $warn->title("Warning");

  my %gui_opts = (-text       => $text,
                  -foreground => 'red',
                  -anchor     => 'c');

  my $warn_label = $warn->Label(%gui_opts)->pack();

  %gui_opts = (-text       => 'Exit',
               -foreground => 'red',
               -command    => sub { exit; });

  my $warn_button = $warn->Button(%gui_opts)->pack();

  $warn->waitWindow;

  return 1;

} # showExitMessage


#############################################################################
#  showWarning
#  -----------

sub showWarning {

  my $text = join ' ', @_;

  my $warn = $gTk->Toplevel();

  $warn->title("Warning");

  my %gui_opts = (-text       => $text,
                  -foreground => 'red',
                  -anchor     => 'c');

  my $warn_label = $warn->Label(%gui_opts)->pack();

  %gui_opts = (-text       => 'Close',
               -foreground => 'red',
               -command    => sub { $warn->withdraw; });

  my $warn_button = $warn->Button(%gui_opts)->pack();

  $warn->waitWindow;

  return 1;

} # showWarning


#############################################################################
#  Die
#  ---

sub Die {

  my $text = shift;

  if ($gOpts{tk} == 1) {

    showExitMessage($text)
      or die $text . "\nCannot show pop-up window.\n";

  } else {

    die $text . "\n";

  }

  return 1;

} # Die


#############################################################################
#  Warn
#  ----

sub Warn {

  my $text = shift;

  if ($gOpts{tk} == 1) {

    showWarning($text)
      or Die($text);

  } else {

    warn $text . "\n";

  }

  return 1;

} # Warn


#############################################################################
#  fileParts
#  ---------

sub fileParts {

  my $file = shift;

  my $disk = '';

  #print "[$gDirSep] [$disk] [$file]\n";

  if ($file =~ /^[A-Z]:/i) {

    ($disk, $file) = $file =~ /^([A-Z]):(.*$)/i;
    $disk .= ':';

  }

  #print "[$gDirSep] [$disk] [$file]\n";

  my @list = split /$gDirSep/, $file;

  my $dir = join $gDirSep, @list[0..$#list-1];

  unless ($dir) {

    $dir  = '.';
    $file = '.' . $gDirSep . $file;

  }

  my $name = $list[-1];

  my @name_list = split /\./, $name;

  my $base = join '.', @name_list[0..$#name_list-1];

  my $ext = $name_list[-1];

  if ($disk =~ /^[A-Z]:$/i) {

    $file = $disk . $file;

  }

  #print "[$file] [$disk] [$dir] [$base] [$ext]\n";

  my $parts = {file => $file,   # Original full file name.
               disk => $disk,   # Windows disk name.
               dir  => $dir,    # Directory part of file (maybe '.').
               base => $base,   # File name without extension.
               ext  => $ext};   # File extension.

  return $parts;

} # fileParts


#############################################################################
#  checkTextFile
#  -------------

sub checkTextFile {

  my $file = shift;

  unless (-f $file) {

    Warn("$file is not a file.");
    return 0;

  }

  unless (-r $file) {

    Warn("$file is not a readable file.");
    return 0;

  }

  unless (-T $file) {

    Warn("$file is not a text file.");
    return 0;

  }

  return 1;

} # checkTextFile


#############################################################################
#  getOptError
#  -----------

sub getOptError {

  my $opt = shift;

  my $text = '';

  if ($gOpts{tk} == 1) {

    my $opt_string = '^#: +\-\-' . $opt . '\s+';

    $text =
      `perl -ne "s/$opt_string//, tr/\n\r.;//d, print if /$opt_string/" $0`;

  } else {

    $text = "Option --${opt}";

  }

    $text .= ": '$gOpts{$opt}'"
      if defined $gOpts{$opt};

  return $text;

} # getOptError


#############################################################################
#  setBoolOpt
#  ----------

sub setBoolOpt {

  my $opt = shift;
  my $def = shift;

  if (exists $gOpts{$opt} and defined $gOpts{$opt}) {

    $gOpts{$opt} = 1
      if $gOpts{$opt} =~ /^[YT]/i;  # Set to 1 if Yes/True.

    $gOpts{$opt} = 0
      if $gOpts{$opt} =~ /^[NF]/i;  # Set to 0 if No/False.

  } else {

    $gOpts{$opt} = $def;

  }

  $gOpts{$opt} = 1                  # Must be True (1) -
    unless $gOpts{$opt} == 0;       #   if not False (0)!

  return 1;

} # setBoolOpt


#############################################################################
#  setNumberOpt
#  ------------

sub setNumberOpt {

  my $opt = shift;
  my $def = shift;

  unless (exists $gOpts{$opt} and defined $gOpts{$opt}) {

    $gOpts{$opt} = $def;

  }

  unless ($gOpts{$opt} =~ /^[\d\.]+$/) {

    my $text = getOptError($opt);

    $text .= ' is not a number.';

    if ($gOpts{tk} == 1) {

      showWarning($text)
        or Die($text);

      return 0;

    } else {

      showUsage($text)
        or Die($text);

      return 0;

    }

  }

  return 1;

} # setNumberOpt


#############################################################################
#  setOpts
#  -------

sub setOpts {

  my $opt = 'os';

  my $text = getOptError($opt);

  $text .= ' cannot be set as a boolean value.';

  setBoolOpt($opt, UNIX)
    or Die($text);

  if ($gOpts{$opt} == WINDOWS) {
    $gDirSep = '\\\\';

  } elsif ($gOpts{$opt} == UNIX) {
    $gDirSep = '/';

  } else {
    $gDirSep = '/';

  }

  # Windows separator '\\\\' does not work, so always use '/'.
  $gDirSep = '/';

  #print "[$gOpts{$opt}] [$gDirSep]\n";

  $opt = 'genefile';

  if (exists $gOpts{extgene} and defined $gOpts{extgene} and $gOpts{extgene}) {

    if (exists $gOpts{$opt} and defined $gOpts{$opt}) {

      # It has been specified by the user.

    } else {

      # Use default gene list file.

      #my $file_base = $0;

      #$file_base =~ s/^([^$gDirSep]*)$gDirSep.*$/$1/;

      #$gOpts{$opt} = $file_base . $gDirSep . DEF_GENE_LIST_FILE;

      #if ($gOpts{os} == WINDOWS) {

      #  $gOpts{$opt} =~ s/\//\\/g;

      #}

      my $file_parts = fileParts($0);

      $gOpts{$opt} = $file_parts->{disk} .
                    $file_parts->{dir}  . $gDirSep .
                    DEF_GENE_LIST_FILE;

      #print "[$gOpts{$opt}]\n";

    }

    checkTextFile($gOpts{$opt})
      or Die("Cannot check file: $gOpts{$opt}.");

  }

  $opt = 'extgene';

  setBoolOpt($opt, DEF_EXT_GENE_LIST)
    or Die($text);

  $opt = 'diamcut';

  if (exists $gOpts{$opt} and defined $gOpts{$opt}) {

    if ($gOpts{$opt} =~ /^[YTNF]/i) {

      my $text = getOptError($opt);

      $text .= ' cannot be set as a boolean value.';

      setBoolOpt($opt, DEF_DIAM_CUT_OPT)
        or Die($text);

      $gOpts{$opt} = DEF_DIAM_CUT_SIZE
        if $gOpts{$opt} == 1;

    } else {

      my $text = getOptError($opt);

      $text .= ' is not a number.';

      setNumberOpt($opt, DEF_DIAM_CUT_SIZE)
        or Die($text);

    }

  } else {

    $gOpts{$opt} = DEF_DIAM_CUT_OPT;

    $gOpts{$opt} = DEF_DIAM_CUT_SIZE
      if $gOpts{$opt} == 1;

  }

  my @opts = qw(bactcont
                window
                maxwin
                blocksize
                blockstart
                bgcut
                upperbg
                lowerbg
                scaling
                spotnorm
                genenorm
                genestats
                alldata
                log
                info);

  my $bool = \&setBoolOpt;
  my $num  = \&setNumberOpt;

  my %opt_set = (bactcont   => $bool,
                 window     => $num,
                 maxwin     => $num,
                 blocksize  => $num,
                 blockstart => $num,
                 bgcut      => $bool,
                 upperbg    => $num,
                 lowerbg    => $num,
                 scaling    => $num,
                 spotnorm   => $bool,
                 genenorm   => $bool,
                 genestats  => $bool,
                 alldata    => $bool,
                 log        => $bool,
                 info       => $bool);

  my %opt_def = (bactcont   => DEF_BACT_CONT_OPT,
                 window     => DEF_WINDOW_SIZE,
                 maxwin     => DEF_MAXWIN_SIZE,
                 blocksize  => DEF_BLOCK_SIZE,
                 blockstart => DEF_BLOCKSTART_SIZE,
                 bgcut      => DEF_BG_CUT_OPT,
                 upperbg    => DEF_UPPER_BG_CUT_SIZE,
                 lowerbg    => DEF_LOWER_BG_CUT_SIZE,
                 scaling    => DEF_SCALE_FACTOR,
                 spotnorm   => DEF_SPOT_NORM_OPT,
                 genenorm   => DEF_GENE_NORM_OPT,
                 genestats  => DEF_GENE_STAT_OPT,
                 alldata    => DEF_ALL_DATA_OPT,
                 log        => DEF_LOG_FILE_OPT,
                 info       => DEF_INFO_FILE_OPT);

 OPT:
  foreach my $opt (@opts) {

    my $text = getOptError($opt);

    $text .= ' is an invalid value for this option.';

    $opt_set{$opt}->($opt, $opt_def{$opt})
      or Die(getOptError($opt) . ' is an invalid value for this option.');

  }

  return 1;

} # setOpts


#############################################################################
#  trimLine
#  --------

sub trimLine {

  my $line = shift;

  chomp $line;

  $line =~ tr/\"\r//d;

  #my @array = split /\t/, $line;

  #s/^\s+//, s/\s+$// foreach @array;

  #$line = join "\t", @array;

  return $line;

} # trimLine


#############################################################################
#  readGeneFile
#  ------------

sub readGeneFile {

  open GENE, $gOpts{genefile}
    or Die("Cannot open file $gOpts{genefile}: $!");

  while (defined (my $line = <GENE>) ) {

    push @gGeneList, trimLine($line);

  }

  close GENE
    or Die("Cannot close file $gOpts{genefile}: $!");

  return 1;

} # readGeneFile


#############################################################################
#  initialReadFile
#  ---------------

sub initialReadFile {

  my $fh = shift;

  my $file = $gFileParts->{file};

  open FILE, $file
    or Die("Cannot open file $file: $!");

  my $gene_flag = 0;

  # Only need a gene list if gene averaged data required.

  if ($gOpts{genenorm} == 1) {

    unless ($gOpts{extgene}) {

      undef @gGeneList;

      my $string = "Using gene list from genes in file: $file.\n\n";

      print $fh $string;

      print STDERR $string
        if $gOpts{log} == 0;

      $gene_flag = 1;

    } else {

      my $string = "Using gene list file: $gOpts{genefile}.\n\n";

      print $fh $string;

      print STDERR $string
        if $gOpts{log} == 0;
    }

  }

  my $header_flag = 0;
  my $old_col     = 0;
  my $old_row     = 0;
  my $max_x       = 0;

  my $count = 0;

 WHILE:
  while (defined (my $line = <FILE>) ) {

    $line = trimLine($line);

    $header_flag = 1, next WHILE
      if $line =~ /^Block/;

    next WHILE
      unless $header_flag == 1;

    $count++;

    my @array = split /\t/, $line;

    my $block = $array[0];
    my $col   = $array[1];
    my $row   = $array[2];
    my $x_pos = $array[5];

    if ($gene_flag == 1) {

      my $id = $array[4];

      $id =~ s/\s+$//;

      push @gGeneList, $id;

    }

    #print "COL: $col\t$old_col\t$gColsBlock\n";

    if ($col > $old_col) {

      $old_col = $col;

    } elsif ($col < $old_col) {

      $gColsBlock = $old_col
        if $gColsBlock == 0;

    } elsif ($gColsBlock and ($col > $gColsBlock or $old_col > $gColsBlock) ) {

      Die("The number of columns per block changes in the GenePix file.");

    }

    #print "ROW: $row\t$old_row\t$gRowsBlock\n";

    if ($row > $old_row) {

      $old_row = $row;

    } elsif ($row < $old_row) {

      $gRowsBlock = $old_row
        if $gRowsBlock == 0;

    } elsif ($gRowsBlock and ($row > $gRowsBlock or $old_row > $gRowsBlock) ) {

      Die("The number of rows per block changes in the GenePix file.");

    }

    if ($col == $gColsBlock and $row == $gRowsBlock) {

      #print "X: $x_pos\t$max_x\t$gBlocksArray\n";

      if ($x_pos > $max_x) {

        $max_x = $x_pos;

      } elsif ($x_pos < $max_x) {

        #print "$block\n";

        $gBlocksArray = $block - 1
          if $gBlocksArray == 0;

        #my $mod = ($block - 1)%$gBlocksArray;

        #print "[$mod]\n";

        $max_x = 0;

      } elsif ($gBlocksArray and ( ($block - 1)%$gBlocksArray != 0) ) {

        Die("The number of blocks per array changes in the GenePix file.");

      }

    }

  }

  my $string =
    "Analysis performed without a dongle in file $file.\n" .
    "Please re-analyse with a dongle in your computer.\n";

  print $fh $string;

  Die($string)
    if $count == 100 or $count == 1000;

  $string = "Cannot calculate columns per block in file $file.\n";

  print $fh $string;

  Die($string)
    if $gColsBlock == 0;

  $string = "Columns per block: " . $gColsBlock . ".\n";

  print $fh $string;

  print STDERR $string
    if $gOpts{log} == 0;

  Die("Cannot calculate rows per block in file $file.\n")
    if $gRowsBlock == 0;

  $string = "Rows per block: " . $gRowsBlock . ".\n";

  print $fh $string;

  print STDERR $string
    if $gOpts{log} == 0;

  Die("Cannot calculate blocks per array in file $file.\n")
    if $gBlocksArray == 0;

  $string = "Blocks per array: " . $gBlocksArray . ".\n\n";

  print $fh $string;

  print STDERR $string
    if $gOpts{log} == 0;

  close FILE
    or Die("Cannot close file $file: $!");

  return 1;

} # initialReadFile


#############################################################################
#  readHeader
#  ----------

sub readHeader {

  my $line = shift;

  if ($line =~ /^Block\s+/) {

    $line =~ s/Rgn R.( ?)/Rgn R2$1/g;   # Replaces unreadable sign with 2.

    $line =~ s/^(Block)\s+/Columns=$1\t/;

  }

  if ($gGPVers == 3) {

    if ($line =~ /^RatioFormulation=/) {

      $line =~ s/^RatioFormulation=/RatioFormulations=/;

    }

  }

  #my $file = $gFileParts->{file};

  # Could check header line is correct here.

  my ($title, @desc) = split /[=\t]+/, $line;

  my $desc = join '; ', @desc;

  $gHeader->{$title} = $desc;

  return 1;

} # readHeader


#############################################################################
#  getHeaderInfo
#  -------------

sub getHeaderInfo {

  my $fh = shift;

  if ($gHeader->{RatioFormulations} =~ m!635( nm)?/532( nm)?!) {

    $gCyRatio = 1;

    print $fh
      "Ratio: 635/532.\n";

  } elsif ($gHeader->{RatioFormulations} =~ m!532( nm)?/635( nm)?!) {

    $gCyRatio = 2;

    print $fh
      "Ratio: 532/635.\n";

  }

  #if ($gHeader->{Creator} =~ /^GenePix Pro 4\./) {

  #  $gGPVers = 4;

  #  print $fh
  #    "File type: GenePix version 4.\n";

  #} elsif ($gHeader->{Creator} =~ /^GenePix Pro 3\./) {

  #  $gGPVers = 3;

  #  print $fh
  #    "File type: GenePix version 3.\n";

  #} else {

  #  $gGPVers = 3;

  #  print $fh
  #    "File type: unknown GenePix version; using default of version 3.\n";

  #}

  if ($gHeader->{Type} =~ /^GenePix Results 3/) {

    $gGPVers = 5;

    print $fh
      "File type: GenePix version 5.\n";


  } elsif ($gHeader->{Type} =~ /^GenePix Results 2/) {

    $gGPVers = 4;

    print $fh
      "File type: GenePix version 4.\n";

  } elsif ($gHeader->{Type} =~ /^GenePix Results 1/) {

    $gGPVers = 3;

    print $fh
      "File type: GenePix version 3.\n";

  } else {

    $gGPVers = 3;

    print $fh
      "File type: unknown GenePix version; using default of version 3.\n";

  }

  print $fh "\n";

  return 1;

} # getHeaderInfo


#############################################################################
#  printHeader
#  -----------

sub printHeader {

  my $file_header = $gFileParts->{disk} .
                    $gFileParts->{dir}  . $gDirSep .
                    $gFileParts->{base} . '_' .
                    'header.txt';

  open HEADER, '>' . $file_header
    or Die("Cannot create file $file_header: $!");

  foreach my $key (sort keys %{$gHeader}) {
    print HEADER $key . ":\t" . $gHeader->{$key} . "\n";
  }

  close HEADER
    or Die("Cannot close file $file_header: $!");

  return 1;

} # printHeader


#############################################################################
#  versGPLines
#  -----------

sub versGPLines {

  my $array = shift;

  my $b635sd2 = 0; # $array[15];
  my $b532sd2 = 0; # $array[24];

  my $f635sat = 0; # $array[16];
  my $f532sat = 0; # $array[25];

  # $ratio is median of ratios, or,
  #   if this is zero, then ratio of medians.

  my $med_ratio = 0;
  my $ratio_med = 0;

  # These are the values of the median backgrounds;
  #   also called 'foreground median' - 'background'.

  my $sig_expt = 0;
  my $sig_cont = 0;

  my $flag = '';

  # Positions of spot variables in the array are different
  #   between versions 3 and 4 of GenePix.

  my $b635sd2_pos = 0;
  my $b532sd2_pos = 0;

  my $f635sat_pos = 0;
  my $f532sat_pos = 0;

  my $med_ratio_pos = 0;
  my $ratio_med_pos = 0;

  my $flag_pos = 0;

  my $sig_a_pos = 0;
  my $sig_b_pos = 0;

  # Could check for correct size of @{$array} here
  #   between different versions of GenePix.

  if ($gGPVers == 3) {

    $b635sd2_pos = 15;
    $b532sd2_pos = 24;

    $f635sat_pos = 16;
    $f532sat_pos = 25;

    $med_ratio_pos = 28;
    $ratio_med_pos = 26;

    $flag_pos = 42;

    $sig_a_pos = 38;    # Cy5.
    $sig_b_pos = 39;    # Cy3.

  } elsif ($gGPVers == 4) {

    $b635sd2_pos = 15;
    $b532sd2_pos = 24;

    $f635sat_pos = 16;
    $f532sat_pos = 25;

    $med_ratio_pos = 46;
    $ratio_med_pos = 44;

    $flag_pos = 80;

    $sig_a_pos = 72;    # Cy5.
    $sig_b_pos = 73;    # Cy3.

  } elsif ($gGPVers == 5) {

    $b635sd2_pos = 18;
    $b532sd2_pos = 30;

    $f635sat_pos = 19;
    $f532sat_pos = 31;

    $med_ratio_pos = 34;
    $ratio_med_pos = 32;

    $flag_pos = 53;

    $sig_a_pos = 45;    # Cy5.
    $sig_b_pos = 46;    # Cy3.

  }

  $b635sd2 = $array->[$b635sd2_pos];
  $b532sd2 = $array->[$b532sd2_pos];

  $f635sat = $array->[$f635sat_pos];
  $f532sat = $array->[$f532sat_pos];

  $med_ratio = $array->[$med_ratio_pos];

  $ratio_med = $array->[$ratio_med_pos];

  $flag = $array->[$flag_pos];

  if ($gCyRatio == 1) {

    $sig_expt = $array->[$sig_a_pos];
    $sig_cont = $array->[$sig_b_pos];

  } elsif ($gCyRatio == 2) {

    $sig_expt = $array->[$sig_b_pos];
    $sig_cont = $array->[$sig_a_pos];

  }

  return ($b635sd2, $b532sd2, $f635sat, $f532sat,
          $med_ratio, $ratio_med, $flag, $sig_expt, $sig_cont);

} # versGPLines


#############################################################################
#  readLine
#  --------

sub readLine {

  my $line = shift;

  my @array = split /\t/, $line;

  my $block = $array[0];

  my $block_col = $array[1];    # Column number within the block.
  my $block_row = $array[2];    # Row number within the block.

  my $mask = $array[3];
  my $id   = $array[4];

  $mask = 'P'
    unless $mask;

  $id =~ s/\s+$//;

  my $x_pos = $array[5];
  my $y_pos = $array[6];

  my $diam = $array[7];

  # Calculate absolute column, row and index position
  #   (within the micro-array).

  my $array_col =
    ( ($block - 1)%$gBlocksArray) * $gColsBlock + $block_col;

  my $array_row =
    (int( ($block - 1)/$gBlocksArray) ) * $gRowsBlock + $block_row;

  # - 1 at the end of the next equation so that array starts at 0, not 1.

  my $index =
    ($array_row - 1) * $gBlocksArray * $gColsBlock + $array_col - 1;

  #print "[$array_col] [$array_row] [$index]\n";

  # These values are in different positions of the GenePix file
  #   for different versions of the software.

  my ($b635sd2, $b532sd2, $f635sat, $f532sat,
      $med_ratio, $ratio_med, $flag, $sig_expt, $sig_cont)
    = versGPLines(\@array);

  my $ratio = $med_ratio;

  $ratio = $ratio_med
    if $ratio == 0;

  my $trust = 0;

  $trust = ($sig_expt < $sig_cont) ? $sig_expt : $sig_cont;

  my $avg_log = 0;

  $avg_log = (log $sig_expt + log $sig_cont)/2
    if ($sig_expt > 0 and $sig_cont > 0);

  my $data = undef;

  $data->[INDEX]       = $index;    # Index location on the micro-array.
  $data->[ARRAY_BLOCK] = $block;    # Block in micro-array.
  $data->[BLOCK_COL]   = $block_col;    # Column in block.
  $data->[BLOCK_ROW]   = $block_row;    # Row in block.
  $data->[ARRAY_COL]   = $array_col;    # Column in micro-array.
  $data->[ARRAY_ROW]   = $array_row;    # Row in micro-array.
  $data->[RATIO]       = $ratio;    # Ratio from GenePix file.
  $data->[NORM_FACTOR] = 0;         # Normalisation factor.
  $data->[NORM_RATIO]  = 0;         # Normalised ratio.
  $data->[X_POS]       = $x_pos;    # X position on micro-array.
  $data->[Y_POS]       = $y_pos;    # Y position on micro-array.
  $data->[MASK]        = $mask;     # Mask (or name) from GenePix file.
  $data->[ID]          = $id;       # ID from GenePix file.
  $data->[FLAG]        = 'A';       # Integrated flag from GenePix file
                                    #   and mask information.
  $data->[GPFLAG]      = $flag;     # GenePix flag.
  $data->[GSFLAG]      = 'A';       # GeneSpring flag.
  $data->[B635SD2]     = $b635sd2;  # % of Cy5 signal pixels > BG 635 + 2SD2.
  $data->[B532SD2]     = $b532sd2;  # % of Cy3 signal pixels > BG 532 + 2SD2.
  $data->[F635SAT]     = $f635sat;  # % of saturated Cy5 signal pixels.
  $data->[F532SAT]     = $f532sat;  # % of saturated Cy3 signal pixels.
  $data->[DIAM]        = $diam;     # Spot diameter.
  $data->[NO_SPOTS]    = 0;         # No. spots to calculate the
                                    #   normalisation factor.
  $data->[MED_RATIO]   = $med_ratio;    # Median of ratios.
  $data->[RATIO_MED]   = $ratio_med;    # Ratio of medians.
  $data->[SIG_EXPT]    = $sig_expt; # Signal1 (experiment).
  $data->[SIG_CONT]    = $sig_cont; # Signal2 (control).
  $data->[TRUST]       = $trust;    # Trust is lowest signal in
                                    #   experiment or control.
  $data->[AVG_LOG_SIG] = $avg_log;  # Average log intensity
                                    #   for both signals.
  $data->[SIG_RATIO]   = 0;         # Ratio of signals.
  $data->[SIG_NORM_FACTOR] = 0;     # Normalisation factor of signals.
  $data->[SIG_NORM_RATIO]  = 0;     # Normalised ratio of signals.
  $data->[NORM_SIG_EXPT]   = 0;     # Normalised signal for experiment.
  $data->[BACT_NORM_RATIO] = 0;     # Normalised ratio (of signals)
                                    #   corrected with bact. controls.

  $gDataArray->[$index] = $data;

  return 1;

} # readLine


#############################################################################
#  readFile
#  --------

sub readFile {

  my $fh = shift;

  my $file = $gFileParts->{file};

  open FILE, $file
    or Die("Cannot open file $file: $!");

  my $header_flag = 0;

  $gHeader    = undef;
  $gDataArray = undef;

 LINE:
  while (defined (my $line = <FILE>) ) {

    $line = trimLine($line);

    if ($header_flag == 0) {

      # $header_flag is set to 1 when at the end of the header.

      $header_flag = 1
        if $line =~ /^Block\s+/;

      readHeader($line)
        or Die("Cannot read header info. for line $line in file $file.");

      if ($header_flag == 1) {

        # At the end of the header, get information about
        #   Cy ratio and GenePix version

        getHeaderInfo($fh)
          or Die("Cannot get useful header information for file $file.");

        if ($gOpts{info} == 1) {

          # Print header information to a file.

          print STDERR "Printing header information to a file.\n\n";

          printHeader()
            or Die("Cannot print header information for file $file.");

        }

      }

      next LINE;

    }

    unless ($header_flag == 1) {
      Die("Cannot read GenePix file $file correctly.");
    }

    # Get relevant data for each spot.

    readLine($line)
      or Die("Cannot read data for line $line in file $file.");

  }

  close FILE
    or Die("Cannot close file $file: $!");

  return 1;

} # readFile


#############################################################################
#  calcDimensions
#  --------------

sub calcDimensions {

  $gBigX = $gColsBlock * $gBlocksArray;

  $gBigY = ($#{$gDataArray} + 1)/$gBigX;

  #print "[$gBigX] [$gBigY]\n";

  return 1;

} # calcDimensions


#############################################################################
#  zeroData
#  --------

sub zeroData {

  my $data  = shift;
  my $index = shift;

  $data->[INDEX]       = $index;
  $data->[ARRAY_BLOCK] = 0;
  $data->[BLOCK_COL]   = 0;
  $data->[BLOCK_ROW]   = 0;
  $data->[ARRAY_COL]   = 0;
  $data->[ARRAY_ROW]   = 0;
  $data->[RATIO]       = 0;
  $data->[NORM_FACTOR] = 0;
  $data->[NORM_RATIO]  = 0;
  $data->[X_POS]       = 0;
  $data->[Y_POS]       = 0;
  $data->[MASK]        = 'A';
  $data->[ID]          = '';
  $data->[FLAG]        = 'A';
  $data->[GPFLAG]      = 0;
  $data->[GSFLAG]      = 'A';
  $data->[B635SD2]     = 0;
  $data->[B532SD2]     = 0;
  $data->[F635SAT]     = 0;
  $data->[F532SAT]     = 0;
  $data->[DIAM]        = 0;
  $data->[NO_SPOTS]    = 0;
  $data->[MED_RATIO]   = 0;
  $data->[RATIO_MED]   = 0;
  $data->[SIG_EXPT]    = 0;
  $data->[SIG_CONT]    = 0;
  $data->[TRUST]       = 0;
  $data->[AVG_LOG_SIG] = 0;
  $data->[SIG_RATIO]   = 0;
  $data->[SIG_NORM_FACTOR] = 0;
  $data->[SIG_NORM_RATIO]  = 0;
  $data->[NORM_SIG_EXPT]   = 0;
  $data->[BACT_NORM_RATIO] = 0;
  
  return $data;

} # zeroData


#############################################################################
#  maskSpots
#  ---------

sub maskSpots {

  my $fh = shift;

  my $count = 0;

  my $bad_spots       = 0;
  my $masked_spots    = 0;
  my $not_found_spots = 0;
  my $marginal_spots  = 0;

 MASK:
  foreach my $index (@{$gDataArray}) {

    unless (defined $index) {

      $index = zeroData($index, $count);

    }

    $count++;

    my $id     = $index->[ID];
    my $mask   = $index->[MASK];
    my $gpflag = $index->[GPFLAG];
    my $flag   = \$index->[FLAG];
    my $ratio  = $index->[RATIO];

    # Set all masked genes to have flags equal to 'A'.

    if ($mask eq 'A') {         # PCR failures or bad spots.

      ${$flag} = 'A';

      $masked_spots++;
      next MASK;

    } elsif ($mask eq 'T') {    # Tags or markers.

      ${$flag} = 'P';

    } elsif ($mask eq 'Y') {    # Budding yeast genes.

      ${$flag} = 'P';

    } elsif ($mask eq 'Q') {    # Bacterial control spots.

      ${$flag} = 'P';

    } elsif ($mask eq 'empty') {

      ${$flag} = 'A';

      $masked_spots++;
      next MASK;

    } elsif ($mask eq 'M1') {

      ${$flag} = 'A';

      $masked_spots++;
      next MASK;

    } elsif (not $mask) {

      ${$flag} = 'A';

      $masked_spots++;
      next MASK;

    }

    # Set all 'negative'/'empty' genes to have flags equal to 'A'.

    if ($id eq 'negative') {

      ${$flag} = 'A';

      $masked_spots++;
      next MASK;

    } elsif ($id eq 'empty') {

      ${$flag} = 'A';

      $masked_spots++;
      next MASK;

    }

    # Set all 'not found'/'error' genes to have flags equal to 'A'.

    if ($gpflag >= 0) {

      ${$flag} = 'P';

    } elsif ($gpflag == -50) {

      ${$flag} = 'A';

      $not_found_spots++;
      next MASK;

    } elsif ($gpflag == -75) {

      ${$flag} = 'A';

      $bad_spots++;
      next MASK;

    } elsif ($gpflag == -100) {

      ${$flag} = 'A';

      $bad_spots++;
      next MASK;

    }

    if ($ratio == 0) {

      ${$flag} = 'A';

      next MASK;

    }

    # Set all flags of marginal genes equal to 'M'.

    if ( ($mask eq 'M') and ($gpflag == 0) ) {

      ${$flag} = 'M';

      $marginal_spots++;
      next MASK;

    }

  }

  print $fh
    "Masked spot information:\n\n" .
    "Number of masked spots:\t\t\t"  . $masked_spots    . ".\n" .
    "Number of not found spots:\t\t" . $not_found_spots . ".\n" .
    "Number of spots flagged bad by user:\t" . $bad_spots . ".\n" .
    "Number of marginal spots:\t\t"  . $marginal_spots  . ".\n\n";

  return 1;

} # maskSpots


#############################################################################
#  bgCutOff
#  --------

sub bgCutOff {

  my $fh = shift;

  # Set flags to 'A' for genes that do not have pixel intensities
  #   above background.

  # Also set flags to 'A' for genes with very low or negative signals;
  #   commented out at the moment.

  my $count = 0;

  my $upper_bg_cut = $gOpts{upperbg};
  my $lower_bg_cut = $gOpts{lowerbg};

  #my $sig_cut = DEF_SIGNAL_CUT;

 BG:
  foreach my $index (@{$gDataArray}) {

    my $flag = \$index->[FLAG];
    my $mask = $index->[MASK];

    my $b635sd2 = $index->[B635SD2];
    my $b532sd2 = $index->[B532SD2];

    my $sig_expt = $index->[SIG_EXPT];
    my $sig_cont = $index->[SIG_CONT];

    next BG
      if $mask eq 'A';

    if ($b635sd2 < $upper_bg_cut and $b532sd2 < $upper_bg_cut) {

      if ($b635sd2 < $lower_bg_cut or $b532sd2 < $lower_bg_cut) {

        #my $id = $index->[ID];
        #print STDERR "[$id]\n";

        ${$flag} = 'A';
        $count++;
        next BG;

      }

    }

    #if ($sig_expt <= $sig_cut or $sig_cont <= $sig_cut) {

    #  ${$flag} = 'A';
    #  next BG;

    #}

  }

  print $fh
    "Using upper background cut-off:\t\t" . $upper_bg_cut . ".\n" .
    "Using lower background cut-off:\t\t" . $lower_bg_cut . ".\n\n" .
    "Number of spots discarded:\t\t" . $count . ".\n\n";

  return 1;

} # bgCutOff


#############################################################################
#  diamCutOff
#  ----------

sub diamCutOff {

  my $fh = shift;

  # Set flags to 'A' for spots that have a diameter
  #   lower than a set cut-off.

  my $count = 0;

  my $diam_cut = $gOpts{diamcut};

 DIAM:
  foreach my $index (@{$gDataArray}) {

    my $flag = \$index->[FLAG];
    my $mask = $index->[MASK];

    my $diam = $index->[DIAM];

    next DIAM
      if $mask eq 'A';

    if ($diam < $diam_cut) {

      ${$flag} = 'A';
      $count++;

    }

  }

  print $fh
    "Using minimum diameter cut-off:\t\t" . $diam_cut . ".\n\n" .
    "Number of spots discarded:\t\t" . $count . ".\n\n";

  return 1;

} # diamCutOff


#############################################################################
#  bactMean
#  --------

sub bactMean {

  my $fh = shift;

  my $bact_no     = 0;
  my $total_ratio = 0;

  # These used to be 'DAP 1:2' etc.

  #my @bact_genes = ('DAP',
  #                  'PHE',
  #                  'THR',
  #                  'TRP',
  #                  'LYS');

 DATA:
  foreach my $index (@{$gDataArray}) {

    my $id   = $index->[ID];
    my $mask = $index->[MASK];
    my $flag = $index->[FLAG];

    #print STDERR "[$mask]\n";

    next DATA
      unless $mask eq 'Q';

    next DATA
      unless $flag eq 'P';

    $bact_no++;

    $total_ratio += $index->[RATIO];

  #BACT:
    #foreach my $bact (@bact_genes) {

    #  if (substr($id, 0, 3) eq $bact) {

    #    ${$flag} = 'A';

    #    $bact_no++;

    #    $total_ratio += $index->[RATIO];

    #    next DATA;

    #  }

    #}

  }

  my $bact_min = DEF_BACT_MIN;

  if ($gCyRatio == 2) {

    $total_ratio = $total_ratio;

  }

  if ($bact_no > $bact_min) {

    $gBactMean = $total_ratio/$bact_no;

    my $bact_mean = sprintf "%.5g", $gBactMean;

    my $string =
      'Mean ratio of control RNAs (expt/control) is ' .
      $bact_mean . ".\n";

    print STDERR $string
      if $gOpts{log} == 0;

    print $fh
      $string . 'Calculated from ' . $bact_no . " spots.\n";

  } else {

    my $string = "Ratio of control RNAs could not be calculated.\n";

    print STDERR $string
      if $gOpts{log} == 0;

    print $fh
      $string . 'There are only ' . $bact_no . " bacterial spots.\n";

  }

  print $fh "\n";

  print STDERR "\n"
    if $gOpts{log} == 0;

  return 1;

} # bactMean


#############################################################################
#  makeGS
#  ------

sub makeGS {

 FLAGS:
  foreach my $index (@{$gDataArray}) {

    my $flag   = $index->[FLAG];
    my $gsflag = \$index->[GSFLAG];

    if ($flag eq 'P') {
      ${$gsflag} = 'P';

    } elsif ($flag eq 'M') {
      ${$gsflag} = 'M';

    } elsif ($flag eq 'A') {
      ${$gsflag} = 'A';

    } else {
      #print "[$flag]\n";
      ${$gsflag} = '-';

    }

  }

  return 1;

} # makeGS


#############################################################################
#  expandWindow
#  ------------

sub expandWindow {

  my $left   = shift;
  my $right  = shift;
  my $top    = shift;
  my $bottom = shift;

  my $array_col = shift;
  my $array_row = shift;

  my $maxwin_size = $gOpts{maxwin};

  if ($left == 1 and $right - $array_col <= $maxwin_size) {
    $right++;
  }

  if ($top == 1 and $bottom - $array_row <= $maxwin_size) {
    $bottom++;
  }

  if ($right == $gBigX and $array_col - $left <= $maxwin_size) {
    $left--;
  }

  if ($bottom == $gBigY and $array_row - $top <= $maxwin_size) {
    $top--;
  }

  return ($left, $right, $top, $bottom);

} # expandWindow


#############################################################################
#  getCount
#  --------

sub getCount {

  my $left   = shift;
  my $right  = shift;
  my $top    = shift;
  my $bottom = shift;

  my $count = 0;

 FORI:
  for (my $i = $left; $i <= $right; $i++) {

  FORJ:
    for (my $j = $top; $j <= $bottom; $j++) {

      my $pos = ($j - 1) * $gBigX + $i - 1;

      my $index = $gDataArray->[$pos];

      my $flag = $index->[FLAG];
      my $mask = $index->[MASK];

      next FORJ
        unless $flag eq 'P';

      next FORJ
        if $mask eq 'T' or $mask eq 'Y' or $mask eq 'Q';

      $count++;

    }

  }

  return $count;

} # getCount


#############################################################################
#  increaseSquare
#  --------------

sub increaseSquare {

  my $left   = shift;
  my $right  = shift;
  my $top    = shift;
  my $bottom = shift;

  my $array_col = shift;
  my $array_row = shift;

  my $block_size = $gOpts{blocksize};

  my $count     = 0;
  my $old_count = 0;
  my $max_flag  = 0;

 INCREASE:
  while (1) {

    ($left, $right, $top, $bottom) =
      expandWindow($left, $right, $top, $bottom, $array_col, $array_row);

    $count =
      getCount($left, $right, $top, $bottom);

    #print "[$count] [$old_count] [$block_size]\n";

    last INCREASE
      if $count > $block_size;

    $max_flag = 1, last INCREASE
      if $count == $old_count;

    $old_count = $count;

  }

  return ($left, $right, $top, $bottom, $max_flag);

} # increaseSquare


#############################################################################
#  getSquareValues
#  ---------------

sub getSquareValues {

  my $left   = shift;
  my $right  = shift;
  my $top    = shift;
  my $bottom = shift;

  my $array_col = shift;
  my $array_row = shift;

  my $ratios = undef;

  my $total_sig_expt = 0;
  my $total_sig_cont = 0;

  # Find positions of spots in a square around a spot (+ and - window);
  #   gets ratio for each spot, and saves them into a list.

  my $count = 0;

 FORX:
  for (my $i = $left; $i <= $right; $i++) {

  FORY:
    for (my $j = $top; $j <= $bottom; $j++) {

      my $pos = ($j - 1) * $gBigX + $i - 1;

      my $index = $gDataArray->[$pos];

      my $flag = $index->[FLAG];
      my $mask = $index->[MASK];

      #print "[$flag] [$mask]\n";

      next FORY
        unless $flag eq 'P';

      next FORY
        if $mask eq 'T' or $mask eq 'Y' or $mask eq 'Q';

      my $ratio = $index->[RATIO];

      push @{$ratios}, $ratio;

      $total_sig_expt += $index->[SIG_EXPT];
      $total_sig_cont += $index->[SIG_CONT];

      $count++;

    }

  }

  my $block_size = $gOpts{blocksize};

  my $max_flag = 0;

  # $count is number of present spots in the square.

  # If less than this,
  #  the square is expanded until a maximum window size is reached.

  unless ($count > $block_size) {

    ($left, $right, $top, $bottom, $max_flag) =
      increaseSquare($left, $right, $top, $bottom, $array_col, $array_row);

    if ($max_flag == 0) {

      # This code is done if the maximum window size has not been reached,
      #   and the square has increased in size.

      $ratios = undef;

      $total_sig_expt = 0;
      $total_sig_cont = 0;

      # Note that recursion is being used here.

      ($ratios, $total_sig_expt, $total_sig_cont) =
        getSquareValues($left, $right, $top, $bottom,
                        $array_col, $array_row);

    }

  }

  return ($ratios, $total_sig_expt, $total_sig_cont);

} # getSquareValues


#############################################################################
#  checkSquareSize
#  ---------------

sub checkSquareSize {

  my $left   = shift;
  my $right  = shift;
  my $top    = shift;
  my $bottom = shift;

  my $array_col = shift;
  my $array_row = shift;

  # Increase block size by 1.5 in size to account
  #   for absent spots in the square.

  # This is a heuristic to speed up the search!

  my $initial_block_size = $gOpts{blockstart};

  my $count     = 0;
  my $old_count = 0;

 COUNT:
  while (1) {

    # Doesn't check for present spots, this is done later.

    $count = ($bottom - $top + 1) * ($right - $left + 1);

    last COUNT
      if $count > $initial_block_size;

    last COUNT
      if $count == $old_count;

    $old_count = $count;

    ($left, $right, $top, $bottom) =
      expandWindow($left, $right, $top, $bottom, $array_col, $array_row);

  }

  return ($left, $right, $top, $bottom);

} # checkSquareSize


#############################################################################
#  defineSquare
#  ------------

sub defineSquare {

  my $index = shift;

  my $array_col = $index->[ARRAY_COL];
  my $array_row = $index->[ARRAY_ROW];

  my $window_size = $gOpts{window};

  my $left = $array_col - $window_size;

  $left = 1
    if $left < 1;

  my $top = $array_row - $window_size;

  $top = 1
    if $top < 1;

  my $right = $array_col + $window_size;

  $right = $gBigX
    if $right > $gBigX;

  my $bottom = $array_row + $window_size;

  $bottom = $gBigY
    if $bottom > $gBigY;

  # Make a square that must have a minimum number of spots in it;
  #   expand the square in two directions if less than this minimum.

  # This does not check whether all spots are present,
  #   and may use a larger square than is necessary.

  ($left, $right, $top, $bottom) =
    checkSquareSize($left, $right, $top, $bottom, $array_col, $array_row);

  my $ratios = undef;

  my $total_sig_expt = 0;
  my $total_sig_cont = 0;

  ($ratios, $total_sig_expt, $total_sig_cont) =
    getSquareValues($left, $right, $top, $bottom, $array_col, $array_row);

  my $norm_signal = 0;

  if ($total_sig_cont > 0) {
    $norm_signal = $total_sig_expt/$total_sig_cont;
  }

  return ($norm_signal, $ratios);

} # defineSquare


#############################################################################
#  calcMedian
#  ----------

sub calcMedian {

  my $ratios = shift;

  my $median = 1;

  my @ratios = sort {$a <=> $b} @{$ratios};

  if (@ratios%2) {

    $median = $ratios[@ratios/2];

  } else {

    if ($ratios[@ratios/2] > 0) {

      $median = exp( (log($ratios[@ratios/2-1]) +
                      log($ratios[@ratios/2]) )/2);

    } else {

      $median = $ratios[@ratios/2];

    }

  }

  return $median;

} # calcMedian


#############################################################################
#  localNormalise
#  --------------

sub localNormalise {

  my $fh = shift;

  my $block_size = $gOpts{blocksize};

  my $count = 0;

  my $low_flag = 0;

 NORM:
  foreach my $index (@{$gDataArray}) {

    my $flag = $index->[FLAG];

    # Ignore bad spots, but not controls.

    next NORM
      if $flag eq 'A';

    # Define square to get a local normalisation value.

    # $ratios is an array of ratios for all spots in this square.

    # $norm_signal is the ratio for the total signals in this square.

    my ($norm_signal, $ratios) = defineSquare($index, $fh);

    $count++;

    print STDERR $count . "\r";

    my $median = calcMedian($ratios);

    my $no_spots = $#{$ratios};

    $index->[NORM_FACTOR] = $median;
    $index->[NO_SPOTS]    = $no_spots;

    if ($no_spots < $block_size) {

      if ($low_flag == 0) {

        print $fh
          'Spots using less than ' . $block_size .
          " spots for normalisation:\n\n";

        print $fh
          "ID\tBlock\tColumn\tRow\tNumber of spots used for normalisation\n\n";

      }

      $low_flag = 1;

      my $id = $index->[ID];

      print $fh
        $id . "\t" .
        $index->[ARRAY_BLOCK] . "\t" .
        $index->[BLOCK_COL] . "\t" .
        $index->[BLOCK_ROW] . "\t" .
        $no_spots . "\n";

    }

    unless ($median == 0) {

      $index->[NORM_RATIO] = $index->[RATIO]/$median;

    }

    if ($index->[SIG_CONT] > 0) {

      $index->[SIG_RATIO] = $index->[SIG_EXPT]/$index->[SIG_CONT];

      my $scaling = $gOpts{scaling};

      $norm_signal = $norm_signal/$scaling;

      $index->[SIG_NORM_FACTOR] = $norm_signal;

      if ($index->[SIG_NORM_FACTOR] > 0 and
          $index->[SIG_RATIO] > 0) {

        # Calculate normalised signals ratio;
        #   (signals ratio)/normalisation factor.

        $index->[SIG_NORM_RATIO] =
          $index->[SIG_RATIO]/$index->[SIG_NORM_FACTOR];

      }

      if ($index->[SIG_NORM_FACTOR] > 0 and
          $index->[SIG_EXPT] > 0) {

        $index->[NORM_SIG_EXPT] =
          $index->[SIG_EXPT]/$index->[SIG_NORM_FACTOR];

        # Modify normalised signals ratio using the bact. controls mean -
        #   $gBactMean, which has not been normalised locally.

      }

      if ($index->[SIG_RATIO] > 0) {

        if ($gBactMean) {
          $index->[BACT_NORM_RATIO] = $index->[SIG_RATIO]/$gBactMean;
        } else {
          $index->[BACT_NORM_RATIO] = 0;
        }

      }

    }

  }

  print $fh "\n"
    if $low_flag == 1;

  return 1;

} # localNormalise


#############################################################################
#  initialiseGeneHash
#  ------------------

sub initialiseGeneHash {

  $gGeneHash = undef;

  foreach my $gene (@gGeneList) {

    $gGeneHash->{$gene}[NO_VAL] = 0;    # No. of values for this gene.
    $gGeneHash->{$gene}[MARG]   = 0;    # Flag if this is a marginal gene.
    $gGeneHash->{$gene}[MEAN]   = 0;    # Mean value for this gene.
    $gGeneHash->{$gene}[CV]     = 0;    # % CV for this gene.

  }

  return 1;

} # initialiseGeneHash


#############################################################################
#  checkPresents
#  -------------

sub checkPresents {

 FOR1:
  foreach my $index (@{$gDataArray}) {

    my $id   = $index->[ID];
    my $flag = $index->[FLAG];
    my $mask = $index->[MASK];

    next FOR1
      if $mask eq 'T' or $mask eq 'Y' or $mask eq 'Q';


    unless (exists $gGeneHash->{$id}) {

      delete $gGeneHash->{$id};
      next FOR1;

    }

    if ($flag eq 'P') {

      $gGeneHash->{$id}[NO_VAL]++;

      my $no_val = $gGeneHash->{$id}[NO_VAL];

      if (defined $index->[NORM_RATIO] and $index->[NORM_RATIO] > 0) {
        $gGeneHash->{$id}[$no_val+DATA_START-1] = log $index->[NORM_RATIO];

      } else {
        $gGeneHash->{$id}[$no_val+DATA_START-1] = 0;

      }

    }

  }

  return 1;

} # checkPresents


#############################################################################
#  checkMarginals
#  --------------

sub checkMarginals {

  my $fh = shift;

  print $fh "Genes with only marginal spots:\n\n";

  my $marg_count = 0;

 FOR2:
  foreach my $index (@{$gDataArray}) {

    my $id   = $index->[ID];
    my $flag = $index->[FLAG];
    my $mask = $index->[MASK];

    next FOR2
      if $mask eq 'T' or $mask eq 'Y' or $mask eq 'Q';

    unless (exists $gGeneHash->{$id}) {

      delete $gGeneHash->{$id};
      next FOR2;

    }

    if ($gGeneHash->{$id}[NO_VAL] == 0 or $gGeneHash->{$id}[MARG] > 0) {

      if ($flag eq 'M') {

        $marg_count++, print $fh $id . "\n"
          if $gGeneHash->{$id}[MARG] == 1;

        $gGeneHash->{$id}[NO_VAL]++;
        $gGeneHash->{$id}[MARG]++;

        my $no_val = $gGeneHash->{$id}[NO_VAL];

        if (defined $index->[NORM_RATIO] and $index->[NORM_RATIO] > 0) {
          $gGeneHash->{$id}[$no_val+DATA_START-1] = log $index->[NORM_RATIO];

        } else {
          $gGeneHash->{$id}[$no_val+DATA_START-1] = 0;

        }

      }

    }

  }

  if ($marg_count > 0) {

    print $fh
      "\nThere are " . $marg_count . " marginal genes.\n\n";

  } else {

    print $fh
      "There are no marginal genes.\n\n";

  }

  return 1;

} # checkMarginals


#############################################################################
#  calcGeneMean
#  ------------

sub calcGeneMean {

 FOR3:
  foreach my $id (keys %{$gGeneHash}) {

    my $no_val = $gGeneHash->{$id}[NO_VAL];

    next FOR3
      if $no_val == 0;

    my $total = 0;

    for (my $i = DATA_START; $i <= ($no_val + DATA_START - 1); $i++) {

      $total += $gGeneHash->{$id}[$i];

    }

    my $mean = $total/$no_val;

    $mean = exp $mean;

    $gGeneHash->{$id}[MEAN] = $mean;

  }

  return 1;

} # calcGeneMean


#############################################################################
#  calcGeneCV
#  ----------

sub calcGeneCV {

 FOR4:
  foreach my $id (keys %{$gGeneHash}) {

    my $no_val = $gGeneHash->{$id}[NO_VAL];
    my $mean   = $gGeneHash->{$id}[MEAN];

    next FOR4
      if $no_val == 0;

    if ($no_val == 1) {
      $gGeneHash->{$id}[CV] = 0;   # CV not calculated if only one value.
      next FOR4;
    }

    my $ssd = 0; # Sum of squared differences between values and the mean.

    for (my $i = DATA_START; $i <= ($no_val + DATA_START - 1); $i++) {

      $ssd += (exp($gGeneHash->{$id}[$i]) - $mean)**2;

    }

    my $std_dev = sqrt($ssd/($no_val - 1) );

    my $cv = 100.0 * ($std_dev/$mean);

    $gGeneHash->{$id}[CV] = $cv;

  }

  return 1;

} # calcGeneCV


#############################################################################
#  calcTotals
#  ----------

sub calcTotals {

  my $fh = shift;

  my $total_expt  = 0;
  my $total_cont  = 0;
  my $total_ratio = 0;

 FOR5:
  foreach my $index (@{$gDataArray}) {

    my $flag    = $index->[FLAG];
    my $mask    = $index->[MASK];
    my $sig_expt = $index->[SIG_EXPT];
    my $sig_cont = $index->[SIG_CONT];

    next FOR5
      if $mask eq 'T' or $mask eq 'Y' or $mask eq 'Q';

    if ($flag eq 'P') {

      $total_expt += $sig_expt;
      $total_cont += $sig_cont;

    }

  }

  $total_ratio = sprintf "%.6g", $total_expt/$total_cont;

  print $fh
    "Overall array signals:\n\n" .
    "Total signal experiment:\t"  . $total_expt  . ".\n" .
    "Total signal control:\t\t"   . $total_cont  . ".\n" .
    "Ratio of total signals:\t\t" . $total_ratio . ".\n\n";

  return 1;

} # calcTotals


#############################################################################
#  calcHighCVs
#  -----------

sub calcHighCVs {

  my $fh = shift;

  my $high_cvs = shift;

 FOR6:
  foreach my $id (sort {lc $a cmp lc $b} @{$high_cvs}) {

    my $no_val = $gGeneHash->{$id}[NO_VAL];
    my $cv     = $gGeneHash->{$id}[CV];

    undef my @list;

    for (my $i = DATA_START; $i <= ($no_val + DATA_START - 1); $i++) {

      push @list, $gGeneHash->{$id}[$i];

    }

    @list = sort {$a <=> $b} @list;

    my $max_ratio = exp($list[$no_val-1])/exp($list[0]);

    printf $fh "%s\t%d\t%.5g\t%.5g\n", $id, $no_val, $cv, $max_ratio;

  }

  print $fh "\n";

  return 1;

} # calcHighCVs


#############################################################################
#  calcQCforReplicates
#  -------------------

sub calcQCforReplicates {

  my $fh = shift;

  my $zero  = 0;
  my $once  = 0;
  my $twice = 0;
  my $more  = 0;

  my $high_cvs = undef;

  my $count = 0;
  my $total = 0;

 FOR7:
  foreach my $id (keys %{$gGeneHash}) {

    my $no_val = $gGeneHash->{$id}[NO_VAL];
    my $cv     = $gGeneHash->{$id}[CV];

    #print "[$id]\n";
    #  if $no_val == 0

  SWITCH: {

      $zero++,  last SWITCH
        if $no_val == 0;

      $once++,  last SWITCH
        if $no_val == 1;

      $twice++, last SWITCH
        if $no_val == 2;

      $more++,  last SWITCH
        if $no_val >= 3;

    }

    # Make list of genes with high CVs.

    push @{$high_cvs}, $id
      if $cv >= DEF_CV_CUT;

    # Used to calculate mean CV.

    if ($no_val >= 1) {

      $count++;
      $total += $cv;

    }

  }

  my $mean_cv = 0;

  if ($count > 0) {
    $mean_cv = sprintf "%.5g", $total/$count;
  } else {
    $mean_cv = 0;
  }

  print $fh
    "Replicate spot information:\n\n" .

    "Zero spots:\t\t" . $zero  . ".\n" .
    "One spot:\t\t"   . $once  . ".\n" .
    "Two spots:\t\t"  . $twice . ".\n" .

    "More than two spots:\t" . $more . ".\n\n" .

    "Average % CV:\t\t" . $mean_cv . ".\n\n" .

    'Genes with CVs higher than ' .
    DEF_CV_CUT . "%:\n\n" .
    "Gene\tSpots\t% CV\tMax. ratio\n\n";

  my $file = $gFileParts->{file};

  if (defined $high_cvs) {

    calcHighCVs($fh, $high_cvs)
      or Die("Cannot calculate genes with high CVs in $file.");

  } else {

    print $fh
      'No genes with CVs higher than ' . DEF_CV_CUT . "%:\n\n";

  }

  return 1;

} # calcQCforReplicates


#############################################################################
#  printNormGenes
#  --------------

sub printNormGenes {

  my $file_ng = $gFileParts->{disk} .
                $gFileParts->{dir}  . $gDirSep .
                $gFileParts->{base} . '_' .
                'norm_genes.txt';

  open NG, '>' . $file_ng
    or Die("Cannot create file $file_ng: $!");

  print NG
    "Gene name\t" .
    "Mean of ratios";

  if ($gOpts{genestats}) {

    print NG
      "\t% CV of ratios\t" .
      "Number of spots\t" .
      "Ratio Values (spread over multiple columns)\n";

  } else {

    print NG "\n";

  }

 FOR8:
  foreach my $id (sort {lc $a cmp lc $b} keys %{$gGeneHash}) {

    my $no_val = $gGeneHash->{$id}[NO_VAL];
    my $mean   = $gGeneHash->{$id}[MEAN];
    my $cv     = $gGeneHash->{$id}[CV];

    printf NG "%s\t", $id;

    if ($mean > 0) {
      printf NG "%.3f", $mean;
    }

    unless ($gOpts{genestats}) {
      print NG "\n";
      next FOR8
    } else {
      print NG "\t";
    }

    if ($cv > 0) {
      printf NG "%.3f", $cv;
    }
    print NG "\t";

    printf NG "%d", $no_val;

    if ($no_val > 0) {

      print NG "\t";

      for (my $i = DATA_START; $i <= ($no_val + DATA_START - 1); $i++) {

        printf NG "%.5g", exp($gGeneHash->{$id}[$i]);

        unless ($i == ($no_val + DATA_START - 1) ) {

          print NG "\t";

        }

      }

    }

    print NG "\n";

  }

  close NG
    or Die("Cannot closefile $file_ng: $!");

  return 1;

} # printNormGenes


#############################################################################
#  spotsToGenes
#  ------------

sub spotsToGenes {

  my $fh = shift;

  my $file = $gFileParts->{file};

  # Initialise gene list hash.

  initialiseGeneHash()
    or Die("Cannot initialise global gene hash in file $file.");

  # Remove genes that are flagged as absent/marginal.

  checkPresents()
    or Die("Cannot check for present genes in $file.");

  # Check for genes that only have one marginal probe set.

  checkMarginals($fh)
    or Die("Cannot add genes that only have marginal probes in $file.");

  # Calculate mean for each gene.

  calcGeneMean()
    or Die("Cannot calculate means for genes in $file.");

  # Calculate standard deviation and
  #   coefficient of variance (CV) for each gene.

  calcGeneCV()
    or Die("Cannot calculate CVs for genes in $file.");

  # Calculate information about total signals.

  calcTotals($fh)
    or Die("Cannot calculate totals for signals of genes in $file.");

  # Calculate QC information about replicate spots.

  calcQCforReplicates($fh)
    or Die("Cannot calculate QC information for replicate spots in $file.");

  printNormGenes()
    or Die("Cannot print normalised ratios for genes in $file.");

  return 1;

} # spotsToGenes


#############################################################################
#  printNormSpots
#  --------------

sub printNormSpots {

  my $file_ns = $gFileParts->{disk} .
                $gFileParts->{dir}  . $gDirSep .
                $gFileParts->{base} . '_' .
                'norm_spots.txt';

  open NS, '>' . $file_ns
    or Die("Cannot create file $file_ns: $!");

  print NS
    "Index\t" .
    "Array block\t" .
    "Block column\t" .
    "Block row\t" .
    "Array column\t" .
    "Array row\t" .
    "Ratio\t" .
    "Normalisation factor (ratios)\t" .
    "Normalised ratio\t" .
    "Array X Pos\t" .
    "Array Y Pos\t" .
    "ID\t" .
    "Flag\t" .
    "Diameter\t" .
    "% > BG + 2SD for Cy5\t" .
    "% > BG + 2SD for Cy3\t" .
    "% saturated pixels for Cy5\t" .
    "% saturated pixels for Cy3\t" .
    "Number of spots used for normalisation\t" .
    "Mask\t" .
    "Median of ratios\t" .
    "Ratio of medians\t" .
    "Median experiment - BG\t" .
    "Median control - BG\t" .
    "Trust value\t" .
    "Average log intensity\t" .
    "Ratio (calculated from signals)\t" .
    "Normalisation factor (signals)\t" .
    "Normalised ratio (signals)\t" .
    "Normalised signal (experiment)\t" .
    "Normalised ratio (using external controls)\n";

 PRINT:
  foreach my $index (@{$gDataArray}) {

    printf NS "%d\t",   $index->[INDEX];

    unless ($index->[ARRAY_BLOCK]) {

      print NS "\t" x 29;

      print NS "\n";

      next PRINT;

    }

    printf NS "%d\t",   $index->[ARRAY_BLOCK];
    printf NS "%d\t",   $index->[BLOCK_COL];
    printf NS "%d\t",   $index->[BLOCK_ROW];
    printf NS "%d\t",   $index->[ARRAY_COL];
    printf NS "%d\t",   $index->[ARRAY_ROW];
    printf NS "%.5g\t", $index->[RATIO];

    if ($index->[NORM_FACTOR]) {
      printf NS "%.5g", $index->[NORM_FACTOR];
    }
    print NS "\t";

    if ($index->[NORM_RATIO]) {
      printf NS "%.5g", $index->[NORM_RATIO];
    }
    print NS "\t";

    printf NS "%d\t",   $index->[X_POS];
    printf NS "%d\t",   $index->[Y_POS];
    printf NS "%s\t",   $index->[ID];
    printf NS "%s\t",   $index->[FLAG];
    printf NS "%d\t",   $index->[DIAM];
    printf NS "%d\t",   $index->[B635SD2];
    printf NS "%d\t",   $index->[B532SD2];
    printf NS "%d\t",   $index->[F635SAT];
    printf NS "%d\t",   $index->[F532SAT];

    if ($index->[NO_SPOTS]) {
      printf NS "%d",   $index->[NO_SPOTS];
    }
    print NS "\t";

    printf NS "%s\t",   $index->[MASK];
    printf NS "%.5g\t", $index->[MED_RATIO];
    printf NS "%.5g\t", $index->[RATIO_MED];
    printf NS "%d\t",   $index->[SIG_EXPT];
    printf NS "%d\t",   $index->[SIG_CONT];
    printf NS "%d\t",   $index->[TRUST];

    if ($index->[AVG_LOG_SIG]) {
      printf NS "%.5g", $index->[AVG_LOG_SIG];
    }
    print NS "\t";

    if ($index->[SIG_RATIO]) {
      printf NS "%.5g", $index->[SIG_RATIO];
    }
    print NS "\t";

    if ($index->[SIG_NORM_FACTOR]) {
      printf NS "%.5g", $index->[SIG_NORM_FACTOR];
    }
    print NS "\t";

    if ($index->[SIG_NORM_RATIO]) {
      printf NS "%.5g", $index->[SIG_NORM_RATIO];
    }
    print NS "\t";

    if ($index->[NORM_SIG_EXPT]) {
      printf NS "%.5g", $index->[NORM_SIG_EXPT];
    }
    print NS "\t";

    if ($index->[BACT_NORM_RATIO]) {
      printf NS "%.5g", $index->[BACT_NORM_RATIO];
    }

    print NS "\n";

  }

  close NS
    or Die("Cannot close file $file_ns: $!");

  return 1;

} # printNormSpots


#############################################################################
#  indexLine
#  ---------

sub indexLine {

  my $line = shift;

  my @array = split /\t/, $line;

  my $block = $array[0];

  my $block_col = $array[1];    # Column number within the block.
  my $block_row = $array[2];    # Row number within the block.

  # Calculate absolute column, row and index position
  #   (within the micro-array).

  my $array_col =
    ( ($block - 1)%$gBlocksArray) * $gColsBlock + $block_col;

  my $array_row =
    (int( ($block - 1)/$gBlocksArray) ) * $gRowsBlock + $block_row;

  # - 1 at the end of the next equation so that array starts at 0, not 1.

  my $index =
    ($array_row - 1) * $gBlocksArray * $gColsBlock + $array_col - 1;

  return $index;

} # indexLine


#############################################################################
#  printAllData
#  ------------

sub printAllData {

  my $fh = shift;

  my $file_ad = $gFileParts->{disk} .
                $gFileParts->{dir}  . $gDirSep .
                $gFileParts->{base} . '_' .
                'all_data.txt';

  open AD, '>' . $file_ad
    or Die("Cannot create file $file_ad: $!");

  my $file = $gFileParts->{file};

  open FILE, $file
    or Die("Cannot open file $file: $!");

  my $header_flag = 0;

 LINE:
  while (defined (my $line = <FILE>) ) {

    $line = trimLine($line);

    if ($header_flag == 0) {

      # $header_flag is set to 1 when at the end of the header.

      print AD $line;

      if ($line =~ /^Block\s+/) {

        $header_flag = 1;

        print AD "\tLocally Normalised Ratio\tIntegrated Flag\tRatio Normalised using Bacterial Controls\n";

      } else {

        print AD "\n";

      }

      next LINE;

    }

    unless ($header_flag == 1) {
      Die("Cannot read GenePix file $file correctly.");
    }

    my $index = indexLine($line);

    print AD $line . "\t";

    if ($gDataArray->[$index][NORM_RATIO]) {
      printf AD "%.5g", $gDataArray->[$index][NORM_RATIO];
    }
    print AD "\t";

    printf AD "%s\t", $gDataArray->[$index][FLAG];

    if ($gDataArray->[$index][BACT_NORM_RATIO]) {
      printf AD "%.5g", $gDataArray->[$index][BACT_NORM_RATIO];
    }

    print AD "\n";

  }

  close FILE
    or Die("Cannot close file $file: $!");

  close AD
    or Die("Cannot close file $file_ad: $!");

  return 1;

} # printAllData


#############################################################################
#  getTimeDate
#  -----------

sub getTimeDate {

  my @array = localtime(time);

  my %month_hash = (1  => 'Jan',
                    2  => 'Feb',
                    3  => 'Mar',
                    4  => 'Apr',
                    5  => 'May',
                    6  => 'Jun',
                    7  => 'Jul',
                    8  => 'Aug',
                    9  => 'Sep',
                    10 => 'Oct',
                    11 => 'Nov',
                    12 => 'Dec');

  my $time_date = sprintf "%02d:%02d on %02d %03s %02d",
    $array[2], $array[1], $array[3], $month_hash{1+$array[4]},
    1900 + $array[5];

  return $time_date

} # getTimeDate


#############################################################################
#  handleFiles
#  -----------

sub handleFiles {

  my @files = sort {$a cmp $b or $a <=> $b} glob $gOpts{file};

 FILES:
  foreach my $file (@files) {

    Warn("Cannot check file: $file."), next FILES
      unless checkTextFile($file);

    $gFileParts = fileParts($file);

    $file = $gFileParts->{file};

    # Many functions output information into the log file.

    my $file_log = '';

    #local *LOG;

    my $fh = undef;     # $fh is file handle for log file.

    if ($gOpts{log} == 1) {

      $file_log = $gFileParts->{disk} .
                  $gFileParts->{dir}  . $gDirSep .
                  $gFileParts->{base} . '_' .
                  'log.txt';

      open LOG, '>' . $file_log
        or Die("Cannot create file $file_log: $!");

      $fh = \*LOG;

      print STDERR 'Log file: ' . $file_log . "\n\n";

    } elsif ($gOpts{log} == 0) {

      # Only do this code once!

      $gOpts{log} = -1;

      $fh = \*STDERR;

      print $fh "Printing log information to stderr.\n\n";

    } elsif ($gOpts{log} == -1) {

      # Do nothing!

    }

    my $script = (split /$gDirSep/, $0)[-1];

    my $string =
      'Using script: ' . $script .
      ' (v ' . VERSION . ").\n\n";

    print $fh $string;

    print STDERR $string
      if $gOpts{log} == 0;

    my $time_date = getTimeDate();

    $string = "Analysing file: $file at " . $time_date . ".\n\n";

    print $fh $string;

    print STDERR $string
      if $gOpts{log} == 0;

    # Calculate the number of columns per block on the micro-array.

    # Also get a gene list from GenePix file if needed!

    initialReadFile($fh)
      or Die("Cannot pre-parse file $file.");

    # Read data into two global hashes: $gHeader and $gDataArray.

    readFile($fh)
      or Die("Cannot parse file $file.");

    calcDimensions()
      or Die("Cannot calculate dimensions of the micro-array in file $file.");

    # Identify genes that are PCR failures, removed by user, etc.

    print STDERR "Masking bad spots.\n\n";

    maskSpots($fh)
      or Die("Cannot mask spots.");

    # Change flags for weaker spots.

    if ($gOpts{bgcut} == 1) {

      print $fh "Using background SD cut-offs.\n\n";

      bgCutOff($fh)
        or Die("Cannot use background SD cut-offs in file $file.");

    } else {

      print $fh "Not using background SD cut-offs.\n\n";

    }

    # Change flags for spots with diameters less than cut-off.

    unless ($gOpts{diamcut} == 0) {

      print $fh "Using spot diameter cut-off.\n\n";

      diamCutOff($fh)
        or Die("Cannot use diameter cut-off in file $file.");

    } else {

      print $fh "Not using spot diameter cut-off.\n\n";

    }

    # Calculate mean ratio for bacterial controls.

    if ($gOpts{bactcont} == 1) {

      print STDERR "Calculating mean ratio for bacterial controls.\n\n";

      bactMean($fh)
        or Die("Cannot calculate mean ratio for controls in file $file.");

    } else {

      my $string = "Not using bacterial controls.\n\n";

      print $fh $string;

      print STDERR $string
        if $gOpts{log} == 0;

    }

    # Change flags to GeneSpring format.

    print STDERR "Changing flags to GeneSpring format.\n\n";

    makeGS()
      or Die("Cannot change flags to GeneSpring format in file $file.");

    if ($gOpts{spotnorm} == 1 or $gOpts{genenorm} == 1 or $gOpts{alldata} == 1) {

      print STDERR "Doing local normalisation.\n\n";

      print $fh
        "Normalisation parameters:\n\n" .
        "Using normalisation window size of:\t" . $gOpts{window} . ".\n" .
        "Using maximum normalisation window of:\t" . $gOpts{maxwin} . ".\n" .
        "Using normalisation block size of:\t" . $gOpts{blocksize} . ".\n" .
        "Using initial block size of:\t\t" . $gOpts{blockstart} . ".\n\n";

      localNormalise($fh)
        or Die("Cannot normalise data in file $file.");

    } else {

      my $string = "Not doing local normalisation.\n\n";

      print $fh $string;

      print STDERR $string
        if $gOpts{log} == 0;

    }

    #print Dumper $gDataArray;

    if ($gOpts{spotnorm} == 1) {

      print STDERR "Printing normalised data for good spots.\n\n";

      printNormSpots()
        or Die("Cannot print normalised data for file $file.");

    } elsif ($gOpts{bgcut} == 1 or $gOpts{diamcut} == 1) {

      print STDERR "Printing filtered data for good spots.\n\n";

      printNormSpots()
        or Die("Cannot print filtered data for file $file.");

    }

    # Only need a gene list if gene averaged data required.

    if ($gOpts{genenorm} == 1) {

      # Well known statement to make gene list unique.

      @gGeneList =
        sort {lc $a cmp lc $b}
        keys %{ { map { $_, 1 } @gGeneList} };

      print STDERR "Printing normalised data for good genes.\n\n";

      spotsToGenes($fh)
        or Die("Cannot convert spot data to gene data in file $file.");

    }

    if ($gOpts{alldata} == 1) {

      print STDERR "Printing all data for all spots on array.\n\n";

      printAllData($fh)
        or Die("Cannot print all data for all spots in file file.");

    }

    $string = "Finished analysis for $file.\n\n";

    print $fh $string;

    print STDERR $string
      if $gOpts{log} == 0;

    #print Dumper $gGeneHash;

    if ($gOpts{log} == 1) {

      close $fh
        or Die("Cannot close file $file_log: $!");

    }

    setGlobals()
      or Die('Cannot reset global variables.');

  }

  print STDERR "Finished all selected files.\n";

  return 1;

} # handleFiles


#############################################################################
#  startAnal
#  ---------

sub startAnal {

  unless (exists $gOpts{file} and defined $gOpts{file}) {

    my $text = 'No files selected.';

    if ($gOpts{tk} == 1) {

      showWarning($text)
        or Die($text);

      return 0;

    } else {

      $text .= "\nNeed to specify at least one file with --file option.\n";

      showUsage($text)
        or Die($text);

      return 0;

    }

  }

  setOpts()
    or Die('Cannot check options.');

  # Only need a gene list if gene averaged data required.

  if ($gOpts{genenorm} == 1) {

    undef @gGeneList;

    # If gene list file is specified - read once at beginning.

    if ($gOpts{extgene}) {

      #print STDERR "Using gene list file $gOpts{genefile}.\n";

      readGeneFile()
        or Die("Cannot read gene list file $gOpts{genefile}.");

    }

  }

  handleFiles()
    or Die('Cannot process files.');

  return 1;

} # startAnal


#############################################################################
#  doCmd
#  -----

sub doCmd {

  my @opts = qw(help
                os=s
                file=s
                genefile=s
                extgene=s
                bactcont=s
                window=s
                maxwin=s
                blocksize=s
                blockstart=s
                bgcut=s
                upperbg=s
                lowerbg=s
                scaling=s
                diamcut=s
                spotnorm=s
                genenorm=s
                genestats=s
                alldata=s
                log=s
                info=s);

  if (GetOptions(\%gOpts, @opts) == 0 or
      exists $gOpts{help}) {

    showUsage()
      or Die('Cannot show usage.');

    return 0;

  }

  startAnal()
    or Die('Cannot start analysis.');

  return 1;

} # doCmd


#############################################################################
#  openFiles
#  ---------

sub openFiles {

  require Tk::FileSelect;

  my %gui_opts = (-directory        => cwd(),
                  -filter           => '*.gpr',
                  -filelabel        => 'Or enter file name:',
                  -filelistlabel    => 'Files',
                  -dirlabel         => 'Path',
                  -dirlistlabel     => 'Directories',
                  -width            => 30,
                  -verify           => ['-T'],      # Accept only text files.
                  -selectmode       => 'multiple');

  my $fs = $gTk->FileSelect(%gui_opts);

  my @files = $fs->Show;

  $gOpts{file} = join ' ', @files
    if defined $files[0];

  return 1;

} # openFiles


#############################################################################
#  openGeneList
#  ------------

sub openGeneList {

  require Tk::FileSelect;

  my %gui_opts = (-directory        => cwd(),
                  -filter           => '*.txt',
                  -filelabel        => 'Or enter file name:',
                  -filelistlabel    => 'Files',
                  -dirlabel         => 'Path',
                  -dirlistlabel     => 'Directories',
                  -width            => 30,
                  -verify           => ['-T'],      # Accept only text files.
                  -selectmode       => 'single');

  my $fs = $gTk->FileSelect(%gui_opts);

  my @files = $fs->Show;

  $gOpts{genefile} = join ' ', @files
    if defined $files[0];

  return 1;

} # openGeneList


#############################################################################
#  doGui
#  -----

sub doGui {

  $gTk->title('Local Normalisation');

  my %gui_opts = (-text        => 'Local Normalisation for GenePix Results',
                  -foreground  => 'red',
                  -height      => 3,
                  -width       => 3,
                  -borderwidth => 10,
                  -relief      => 'groove');

  my %pack_opts = (-fill => 'both');

  my $header_label = $gTk->Label(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-height      => 10,
               -width       => 20,
               -borderwidth => 10,
               -relief      => 'groove');

  %pack_opts = (-padx => 2,
                -pady => 2,
                -side => 'top');

  my $top_frame = $gTk->Frame(%gui_opts)->pack(%pack_opts);

  %pack_opts = (-padx => 2,
                -pady => 2,
                -side => 'bottom');

  my $bottom_frame = $gTk->Frame(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-height      => 20,
               -width       => 10,
               -borderwidth => 10,
               -relief      => 'groove');

  %pack_opts = (-ipadx => 40,
                -fill  => 'both');

  my $middle_frame = $gTk->Frame(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-height      => 10,
               -width       => 20,
               -borderwidth => 10,
               -relief      => 'groove');

  %pack_opts = (-padx => 2,
                -pady => 2,
                -side => 'left');

  my $left_frame = $middle_frame->Frame(%gui_opts)->pack(%pack_opts);

  %pack_opts = (-padx => 2,
                -pady => 2,
                -side => 'right');

  my $right_frame = $middle_frame->Frame(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text       => 'Files to be processed:',
               -foreground => 'purple');

  %pack_opts = (-padx => 2,
                -pady => 2);

  my $file_label = $top_frame->Label(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text             => 'File selector',
               -foreground       => 'purple',
               -activebackground => 'purple',
               -borderwidth      => 10,
               -relief           => 'groove',
               -command          => \&openFiles);

  my $open_dir_button = $top_frame->Button(%gui_opts)->pack();

  %gui_opts = (-text       => 'Normalisation options:',
               -foreground => 'dark green');

  %pack_opts = (-padx => 2,
                -pady => 2);

  my $norm_label = $left_frame->Label(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text           => 'Use external gene list',
               -foreground     => 'dark green',
               -selectcolor    => 'dark green',
               -highlightcolor => 'dark green',
               -variable       => \$gOpts{extgene},
               -anchor         => 'w');

  my $gene_list_check = $left_frame->Checkbutton(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text             => 'Gene list file',
               -foreground       => 'dark green',
               -activebackground => 'dark green',
               -borderwidth      => 10,
               -relief           => 'groove',
               -command          => \&openGeneList);

  %pack_opts = (-side   => 'top',
                -expand => 1,
                -anchor => 'w');

  my $gene_list_button = $left_frame->Button(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text       => 'Minimum window size:',
               -foreground => 'dark green');

  my $win_label = $left_frame->Label(%gui_opts)->pack(%pack_opts);

  my %text_opts = (-textvariable => \$gOpts{window});

  my $win_size = $left_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text       => 'Maximum window size:',
               -foreground => 'dark green');

  my $maxwin_label = $left_frame->Label(%gui_opts)->pack(%pack_opts);

  %text_opts = (-textvariable => \$gOpts{maxwin});

  my $maxwin_size = $left_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text       => 'Minimum normalisation block size:',
               -foreground => 'dark green');

  my $block_label = $left_frame->Label(%gui_opts)->pack(%pack_opts);

  %text_opts = (-textvariable => \$gOpts{blocksize});

  my $block_size = $left_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text       => "Initial normalisation block size" .
                              "\n (accounts for absent spots):",
               -foreground => 'dark green');

  my $blockstart_label = $left_frame->Label(%gui_opts)->pack(%pack_opts);

  %text_opts = (-textvariable => \$gOpts{blockstart});

  my $blockstart_size = $left_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text           => 'Spot normalisation file',
               -foreground     => 'dark green',
               -selectcolor    => 'dark green',
               -highlightcolor => 'dark green',
               -variable       => \$gOpts{spotnorm},
               -anchor         => 'w');

  my $spot_check = $left_frame->Checkbutton(%gui_opts)->pack(%pack_opts);

  #$spot_check->select();

  %gui_opts = (-text           => 'Gene normalisation file',
               -foreground     => 'dark green',
               -selectcolor    => 'dark green',
               -highlightcolor => 'dark green',
               -variable       => \$gOpts{genenorm},
               -anchor         => 'w');

  my $sig_check = $left_frame->Checkbutton(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text           => 'Print all data to new file',
               -foreground     => 'dark green',
               -selectcolor    => 'dark green',
               -highlightcolor => 'dark green',
               -variable       => \$gOpts{alldata},
               -anchor         => 'w');

  my $all_data_check = $left_frame->Checkbutton(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text           => "Normalise using\n  bacterial controls",
               -foreground     => 'dark green',
               -selectcolor    => 'dark green',
               -highlightcolor => 'dark green',
               -variable       => \$gOpts{bactcont},
               -anchor         => 'w');

  my $bact_check = $left_frame->Checkbutton(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text       => "Scale Factor (when expt/control\n" .
                              " channel signals differ):",
               -foreground => 'dark green');

  my $scaling_label = $left_frame->Label(%gui_opts)->pack(%pack_opts);

  %text_opts = (-textvariable => \$gOpts{scaling});

  my $scaling_value = $left_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text       => 'Filtering options:',
               -foreground => 'dark green');

  %pack_opts = (-padx => 2,
                -pady => 2);

  my $filt_label = $right_frame->Label(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text           => 'Background SD cut-offs',
               -foreground     => 'dark green',
               -selectcolor    => 'dark green',
               -highlightcolor => 'dark green',
               -variable       => \$gOpts{bgcut},
               -anchor         => 'w');

  %pack_opts = (-side   => 'top',
                -expand => 1,
                -anchor => 'w');

  my $bg_cut_offs = $right_frame->Checkbutton(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text       => 'Lower cut-off:',
               -foreground => 'dark green');

  my $lower_label = $right_frame->Label(%gui_opts)->pack(%pack_opts);

  %text_opts = (-textvariable => \$gOpts{lowerbg});

  my $lower_bg = $right_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text       => 'Upper cut-off:',
               -foreground => 'dark green');

  my $upper_label = $right_frame->Label(%gui_opts)->pack(%pack_opts);

  %text_opts = (-textvariable => \$gOpts{upperbg});

  my $upper_bg = $right_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text       => "Diameter cut-off\n  (0 means no cut-off used):",
               -foreground => 'dark green');

  my $diam_label = $right_frame->Label(%gui_opts)->pack(%pack_opts);

  %text_opts = (-textvariable => \$gOpts{diamcut});

  my $diam_cut = $right_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text             => 'Start analysis',
               -foreground       => 'red',
               -activebackground => 'red',
               -borderwidth      => 10,
               -relief           => 'raised',
               -cursor           => 'mouse',
               -command          => \&startAnal);

  my $start_button = $bottom_frame->Button(%gui_opts)->pack();

  %gui_opts = (-text    => 'Exit',
               -command => sub { exit; },
               -cursor  => 'man');

  %pack_opts = (-side => 'bottom',
                -fill => 'x');

  my $exit_button = $bottom_frame->Button(%gui_opts)->pack(%pack_opts);

  MainLoop;

  return 1;

} # doGui


#############################################################################
#  MAIN:
#  -----

MAIN: {

  undef %gOpts;

  setGlobals()
    or Die('Cannot set global variables.');

  if (@ARGV == 0) {

    $gTk = new MainWindow;

    $gOpts{tk} = 1;

  } else {

    $gTk = undef;

    $gOpts{tk} = 0;

  }

  # UNIX is set as the default OS,
  #   unless $ENV{OS} indicates it is Windows.

  if (exists $ENV{OS} and defined $ENV{OS}) {

    if ($ENV{OS} =~ /^Windows/i) {
      $gOpts{os} = WINDOWS;
    } else {
      $gOpts{os} = UNIX;
    }

  } else {

    $gOpts{os} = UNIX;

  }

  # Set default options before user interface starts.

  setOpts()
    or Die('Cannot set default options.');

  if ($gOpts{tk} == 1) {

    # Tk GUI version.

    doGui()
      or exit 8;

  } else {

    # Command line version.

    doCmd()
      or exit 8;

  }

  exit 0;

} # MAIN:
